
enum Mode {CYCL, EDGE, NONE} ;

class RemapSkeleton : public BlockMessSchedule {
public :
  RemapSkeleton(const DAD* dst, const DAD* src, const int len) ;
private :
  const DAD* rem ;
  const DAD* loc ;

  int rank ;

  int* exts ;
  int* steps ;

  void (RemapSkeleton :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;


class Remap {
public :
  Remap(const DAD* dst, const DAD* src, const int len) :
          skeleton(dst, src, len), bcst(dst, len) {}

  void execute(void* dstDat, void* srcDat) {
    skeleton.gather(dstDat, srcDat) ;
    bcst.execute(dstDat) ;
  }
private :
  RemapSkeleton skeleton ;
  Replicate bcst ;
} ;


class Shift : public BlockMessSchedule {
public :
  Shift(const DAD* dst, const DAD* src, const int len,
        const int shift, const int dim, const Mode mode) ;

  void execute(void* dstDat, void* srcDat) { gather(dstDat, srcDat) ; }
private :
  int rank, dim ;

  int nShfts ;
  int shftExts [2] ;
  int shftSrcDisps [2] ;
  int shftDstDisps [2] ;

  int* exts ;
  int* steps ;

  const DAD* loc ;

  int* shftRemDisps ;
  int* shftLocDisps ;

  void (Shift :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;


class Skew : public BlockMessSchedule {
public :
  Skew(const DAD* _dst, const DAD* _src, const int len,
       const DAD* _shf, int* shfDat,
       const int _dim, const Mode _mode) ;

  void execute(void* dstDat, void* srcDat) { gather(dstDat, srcDat) ; }
private :
  const DAD* dst ;
  const DAD* src ;
  const DAD* shf ;

  int dim ;
  Mode mode ;

  int rank ;

  Group reduced ;

  void loop(const int dstOffset, const int srcOffset, int* shfData,
            const int r, const int s) ;

  void shift(int offset, Stride str,
             const int ext, const int locDisp, const int remDisp,
             void (Skew :: *transfer)(int, int, int, int)) ;

  void send(int offset, int step, int count, int id) ;
  void recv(int offset, int step, int count, int id) ;
} ;


class MultiShift : public BlockMessSchedule {
public :
  MultiShift(const DAD* dst, const DAD* src, const int len,
             int* shift, const Mode* mode) ;

  void execute(void* dstDat, void* srcDat) { gather(dstDat, srcDat) ; }
private :
  int rank ;

  int* nShfts ;
  int (*shftExts) [2] ;
  int (*shftSrcDisps) [2] ;
  int (*shftDstDisps) [2] ;

  int* exts ;
  int* steps ;

  const DAD* loc ;

  int (*shftRemDisps) [2] ;
  int (*shftLocDisps) [2] ;

  void (MultiShift :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `WriteFaces'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class WriteFaces : public BlockMessSchedule {
public :
  WriteFaces(const DAD* _src, const int len,
             const int _dim,
             const int* _wlo, const int* _whi, const Mode* _mode) ;
private :
  const DAD* src ;

  int rank, dim ;
  const int* wlo ;
  const int* whi ;
  const Mode* mode ;

  int* exts ;
  int* steps ;

  void loop(const int offset, const int ctg, const int vol,
            const Group& remGrp, const int r) ;

  void recvLoop(const int offset, const int ctg, const int vol,
                const Group& remGrp, const int r) ;

  void sendLoop(const int offset, const int ctg, const int vol,
                const Group& remGrp, const int r) ;

} ;


class WriteHalo {
public :
  WriteHalo(const DAD* src, const int len,
            const int* wlo, const int* whi, const Mode* mode) :
          rank(src->rnk()) {
    faceSchedules = new WriteFaces* [rank] ;
    for(int r = 0 ; r < rank ; r++)
      if(mode [r] != NONE && (wlo [r] > 0 || whi [r] > 0))
	faceSchedules [r] = new WriteFaces(src, len, r, wlo, whi, mode) ;
      else
	faceSchedules [r] = 0 ;
  }

  ~WriteHalo() {
    for(int r = 0 ; r < rank ; r++)
      if(faceSchedules [r]) delete faceSchedules [r] ;
    delete [] faceSchedules ;
  }

  void execute(void* base) {
    for(int r = 0 ; r < rank ; r++)
      if(faceSchedules [r]) faceSchedules [r]->gather(base, base) ;
  }

  void show() {
    for(int r = 0 ; r < rank ; r++)
      if(faceSchedules [r]) faceSchedules [r]->show() ;
  }
private :
  const int rank ;
  WriteFaces** faceSchedules ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Make `Group' arguments to `loop' into `const Group &'.

